//-----------------------------------------------------------------------------
//           Name: matrix4x4f.h
//         Author: Kevin Harris
//  Last Modified: 02/01/05
//    Description: OpenGL compatible utility class for a 4x4 matrix of floats.
//                 NOTE: This class has been left unoptimized for readability.
//-----------------------------------------------------------------------------

#ifndef _MATRIX4X4F_H_
#define _MATRIX4X4F_H_ 1

#include <windows.h>
#include <gl\gl.h>			
#include <gl\glu.h>			
#include <gl\glaux.h>	
#include <memory>
#include <cmath>
#include <iostream>	
#include "vecteur3X.h"
#include "vecteur4X.h"

using namespace std;

#define DEGTORAD(degree) ((degree) * (3.141592654f / 180.0f))
#define RADTODEG(radian) ((radian) * (180.0f / 3.141592654f))

typedef vecteur3X<float> vecteur3f;
typedef vecteur4X<float> vecteur4f;
typedef vecteur3X<int> vecteur3i;
typedef vecteur4X<int> vecteur4i;


//***********************************************************************************

class matrix4x4f
{
public:

    GLfloat m[16];

    matrix4x4f(const GLfloat src[16]);
    matrix4x4f()  { identity(); }

    matrix4x4f( GLfloat m0, GLfloat m4, GLfloat  m8, GLfloat m12,
                GLfloat m1, GLfloat m5, GLfloat  m9, GLfloat m13,
                GLfloat m2, GLfloat m6, GLfloat m10, GLfloat m14,
                GLfloat m3, GLfloat m7, GLfloat m11, GLfloat m15 );

    void identity(void);
    void translate(const vecteur3f &trans);
    void translate_x(const GLfloat &dist);
    void translate_y(const GLfloat &dist);
    void translate_z(const GLfloat &dist);
    void rotate(const GLfloat &angle, vecteur3f &axis);
    void rotate_x(const GLfloat &angle);
    void rotate_y(const GLfloat &angle);
    void rotate_z(const GLfloat &angle);
    void scale(const vecteur3f &scale);
    void transformPoint( vecteur3f *vec,float coef=1);
    void transformVector( vecteur3f *vec,float coef=1);
	void ortho(vecteur3f* vec,GLint i);
	void trans1(const vecteur3f &vec );
	void trans(const vecteur3f &vec );

    // Static utility methods
    static matrix4x4f matrix4x4f::invertMatrix( const matrix4x4f *matIn );
    static matrix4x4f matrix4x4f::transpose( const matrix4x4f *matIn );
	static void mix(const matrix4x4f*,const matrix4x4f*,GLfloat,matrix4x4f*);
    
    // Operators...
    matrix4x4f &operator = (const matrix4x4f &other);
    matrix4x4f operator + (const matrix4x4f &other);
    matrix4x4f operator - (const matrix4x4f &other);
    matrix4x4f operator * (const matrix4x4f &other);

    matrix4x4f operator * (const GLfloat scalar);
};

//**************matrix4x********************************************************
inline matrix4x4f::matrix4x4f( GLfloat m0, GLfloat m4, GLfloat  m8, GLfloat m12,
                        GLfloat m1, GLfloat m5, GLfloat  m9, GLfloat m13,
                        GLfloat m2, GLfloat m6, GLfloat m10, GLfloat m14,
                        GLfloat m3, GLfloat m7, GLfloat m11, GLfloat m15 )
{
    m[0]=m0; m[1]=m4; m[2] =m8;  m[3]=m12;
    m[4]=m1; m[5]=m5; m[6] =m9;  m[7]=m13;
    m[8]=m2; m[9]=m6; m[10]=m10; m[11]=m14;
    m[12]=m3; m[13]=m7; m[14]=m11; m[15]=m15;
}

inline matrix4x4f::matrix4x4f(const GLfloat src[16])
{
	memcpy( m, src, sizeof( src ) );
}


inline void matrix4x4f::identity( void )
{
    m[0]=1.0f; m[1]=0.0f; m[2] =0.0f; m[3]=0.0f;
    m[4]=0.0f; m[5]=1.0f; m[6] =0.0f; m[7]=0.0f;
    m[8]=0.0f; m[9]=0.0f; m[10]=1.0f; m[11]=0.0f;
    m[12]=0.0f; m[13]=0.0f; m[14]=0.0f; m[15]=1.0f;
}

inline void matrix4x4f::trans( const vecteur3f &trans )
{
	m[3]=m[0]*(trans.getX())+m[1]*(trans.getY())+m[2]*(trans.getZ())+m[3];
	m[7]=m[4]*(trans.getX())+m[5]*(trans.getY())+m[6]*(trans.getZ())+m[7];
	m[11]=m[8]*(trans.getX())+m[9]*(trans.getY())+m[10]*(trans.getZ())+m[11];
};


inline void matrix4x4f::trans1( const vecteur3f &trans )
{
	m[3]=m[0]*(-trans.getX())+m[1]*(-trans.getY())+m[2]*(-trans.getZ())+m[3];
	m[7]=m[4]*(-trans.getX())+m[5]*(-trans.getY())+m[6]*(-trans.getZ())+m[7];
	m[11]=m[8]*(-trans.getX())+m[9]*(-trans.getY())+m[10]*(-trans.getZ())+m[11];
};


inline void matrix4x4f::translate( const vecteur3f &trans )
{
	m[3] = trans.getX();
    m[7] = trans.getY();
    m[11] = trans.getZ();
}

inline void matrix4x4f::translate_x( const GLfloat &dist )
{
    m[3] += dist;
}

inline void matrix4x4f::translate_y( const GLfloat &dist )
{
    m[7] += dist;
}

inline void matrix4x4f::translate_z( const GLfloat &dist )
{
    m[11] += dist;
}

inline void matrix4x4f::rotate( const GLfloat &angle, vecteur3f &axis )
{
    GLfloat s = sin(DEGTORAD(angle));
    GLfloat c = cos(DEGTORAD(angle));

    axis.normaliz();

	GLfloat ux = axis.getX();
    GLfloat uy = axis.getY();
    GLfloat uz = axis.getZ();
    
    m[0]  = c + (1-c) * ux;
    m[4]  = (1-c) * ux*uy + s*uz;
    m[8]  = (1-c) * ux*uz - s*uy;
    
    m[1]  = (1-c) * uy*ux - s*uz;
    m[5]  = c + (1-c) * pow(uy,2);
    m[9]  = (1-c) * uy*uz + s*ux;
    
    m[2]  = (1-c) * uz*ux + s*uy;
    m[6]  = (1-c) * uz*uz - s*ux;
    m[10] = c + (1-c) * pow(uz,2);
}

inline void matrix4x4f::rotate_x( const GLfloat &angle )
{
    GLfloat s = sin(DEGTORAD(angle));
    GLfloat c = cos(DEGTORAD(angle));

    m[5]  =  c;
    m[9]  =  s;
    m[6]  =  -s;
    m[10] =  c;
}

inline void matrix4x4f::rotate_y( const GLfloat &angle )
{
    GLfloat s = sin(DEGTORAD(angle));
    GLfloat c = cos(DEGTORAD(angle));

    m[0]  =  c;
    m[8]  = -s;
    m[2]  =  s;
    m[10] =  c;
}

inline void matrix4x4f::rotate_z( const GLfloat &angle )
{
    GLfloat s = sin(DEGTORAD(angle));
    GLfloat c = cos(DEGTORAD(angle));

    m[0] =  c;
    m[4] =  s;
    m[1] = -s;
    m[5] =  c;
}

inline void matrix4x4f::scale( const vecteur3f &scale )
{
    m[0]  = scale.getX();
    m[5]  = scale.getY();
    m[10] = scale.getZ();
}

inline void matrix4x4f::transformPoint( vecteur3f *vec, float coef)
{
	vec->set((vec->getX() * m[0] + vec->getY() * m[1] + vec->getZ() * m[2] + m[3])*coef,
		  (vec->getX() * m[4] + vec->getY() * m[5] + vec->getZ() * m[6] + m[7])*coef,
		  (vec->getX() * m[8] + vec->getY() * m[9] + vec->getZ() * m[10]+ m[11])*coef);
}

inline void matrix4x4f::transformVector( vecteur3f *vec, float coef)
{
	vec->set((vec->getX() * m[0] + vec->getY() * m[1] + vec->getZ() * m[2])*coef,
		  (vec->getX() * m[4] + vec->getY() * m[5] + vec->getZ() * m[6])*coef,
		  (vec->getX() * m[8] + vec->getY() * m[9] + vec->getZ() * m[10])*coef);
}

inline matrix4x4f matrix4x4f::operator + ( const matrix4x4f &other )
{
	return matrix4x4f( 
		m[0]  + other.m[0],m[4]  + other.m[4],m[8]  + other.m[8],m[12]  + other.m[12],
		m[1]  + other.m[1],m[5]  + other.m[5],m[9]  + other.m[9],m[13]  + other.m[13],
		m[2]  + other.m[2],m[6]  + other.m[6],m[10] + other.m[10],m[14] + other.m[14],
		m[3] + other.m[3],m[7] + other.m[7],m[11] + other.m[11],m[15] + other.m[15]);
}

inline matrix4x4f &matrix4x4f::operator = ( const matrix4x4f &other )
{
	memcpy( m,  other.m, sizeof(  other.m ) );
	return *this;
}

inline matrix4x4f matrix4x4f::operator - ( const matrix4x4f &other )
{
	return matrix4x4f(
		m[0]  - other.m[0],m[4]  - other.m[4],m[8]  - other.m[8],m[12]  - other.m[12],
		m[1]  - other.m[1],m[5]  - other.m[5],m[9]  - other.m[9],m[13]  - other.m[13],
		m[2]  - other.m[2],m[6]  - other.m[6],m[10] - other.m[10],m[14] - other.m[14],
		m[3] - other.m[3],m[7] - other.m[7],m[11] - other.m[11],m[15] - other.m[15]);
}


inline matrix4x4f matrix4x4f::operator * ( const matrix4x4f &other )
{
	return matrix4x4f((m[0]*other.m[0])+(m[1]*other.m[4])+(m[2]*other.m[8])+(m[3]*other.m[12]),
		(m[4]*other.m[0])+(m[5]*other.m[4])+(m[6]*other.m[8])+(m[7]*other.m[12]),
		(m[8]*other.m[0])+(m[9]*other.m[4])+(m[10]*other.m[8])+(m[11]*other.m[12]),
		(m[12]*other.m[0])+(m[13]*other.m[4])+(m[14]*other.m[8])+(m[15]*other.m[12]),
		(m[0]*other.m[1])+(m[1]*other.m[5])+(m[2]*other.m[9])+(m[3]*other.m[13]),
		(m[4]*other.m[1])+(m[5]*other.m[5])+(m[6]*other.m[9])+(m[7]*other.m[13]),
		(m[8]*other.m[1])+(m[9]*other.m[5])+(m[10]*other.m[9])+(m[11]*other.m[13]),
		(m[12]*other.m[1])+(m[13]*other.m[5])+(m[14]*other.m[9])+(m[15]*other.m[13]),
		(m[0]*other.m[2])+(m[1]*other.m[6])+(m[2]*other.m[10])+(m[3]*other.m[14]),
		(m[4]*other.m[2])+(m[5]*other.m[6])+(m[6]*other.m[10])+(m[7]*other.m[14]),
		(m[8]*other.m[2])+(m[9]*other.m[6])+(m[10]*other.m[10])+(m[11]*other.m[14]),
		(m[12]*other.m[2])+(m[13]*other.m[6])+(m[14]*other.m[10])+(m[15]*other.m[14]),
		(m[0]*other.m[3])+(m[1]*other.m[7])+(m[2]*other.m[11])+(m[3]*other.m[15]),
		(m[4]*other.m[3])+(m[5]*other.m[7])+(m[6]*other.m[11])+(m[7]*other.m[15]),
		(m[8]*other.m[3])+(m[9]*other.m[7])+(m[10]*other.m[11])+(m[11]*other.m[15]),
		(m[12]*other.m[3])+(m[13]*other.m[7])+(m[14]*other.m[11])+(m[15]*other.m[15]));
}


inline matrix4x4f matrix4x4f::operator * ( const GLfloat scalar )
{
	return matrix4x4f(
		m[0]  * scalar , m[4]  * scalar , m[8]  * scalar , m[12]  * scalar,
		m[1]  * scalar , m[5]  * scalar , m[9]  * scalar , m[13]  * scalar,
		m[2]  * scalar , m[6]  * scalar , m[10] * scalar , m[14] * scalar,
		m[3] * scalar , m[7] * scalar , m[11] * scalar , m[15] * scalar);
}


inline void matrix4x4f::mix(const matrix4x4f *mat1,const matrix4x4f *mat2,const GLfloat facteur,matrix4x4f *dest ){
	dest->m[0]=(mat2->m[0]-mat1->m[0])*facteur+mat1->m[0];
	dest->m[4]=(mat2->m[4]-mat1->m[4])*facteur+mat1->m[4];
	dest->m[8]=(mat2->m[8]-mat1->m[8])*facteur+mat1->m[8];
	dest->m[12]=(mat2->m[12]-mat1->m[12])*facteur+mat1->m[12];
	dest->m[1]=(mat2->m[1]-mat1->m[1])*facteur+mat1->m[1];
	dest->m[5]=(mat2->m[5]-mat1->m[5])*facteur+mat1->m[5];
	dest->m[9]=(mat2->m[9]-mat1->m[9])*facteur+mat1->m[9];
	dest->m[13]=(mat2->m[13]-mat1->m[13])*facteur+mat1->m[13];
	dest->m[2]=(mat2->m[2]-mat1->m[2])*facteur+mat1->m[2];
	dest->m[6]=(mat2->m[6]-mat1->m[6])*facteur+mat1->m[6];
	dest->m[10]=(mat2->m[10]-mat1->m[10])*facteur+mat1->m[10];
	dest->m[14]=(mat2->m[14]-mat1->m[14])*facteur+mat1->m[14];
	dest->m[3]=(mat2->m[3]-mat1->m[3])*facteur+mat1->m[3];
	dest->m[7]=(mat2->m[7]-mat1->m[7])*facteur+mat1->m[7];
	dest->m[11]=(mat2->m[11]-mat1->m[11])*facteur+mat1->m[11];
	dest->m[15]=(mat2->m[15]-mat1->m[15])*facteur+mat1->m[15];
};

inline void matrix4x4f::ortho(vecteur3f* vec,GLint i)
{
	identity();
	GLfloat ux=(float) i;
    GLfloat uy=0.0;
    GLfloat uz=0.0;
    m[0]  = ux*ux;
    m[1]  = ux*uy + uz;
    m[2]  = ux*uz - uy;
    
    m[4]  = uy*ux - uz;
    m[5]  = uy*uy;
    m[6]  = uy*uz + ux;
    
    m[8]  = uz*ux + uy;
    m[9]  = uy*uz - ux;
    m[10] = uz*uz;
	transformPoint(vec);
}


#endif // _MATRIX4X4F_H_

